/* irmix -- an OSS mixer controlled by lirc.
 * Copyright (C) 2001 Alastair M. Robinson
 * Loosely based on wmix, by timecop@japan.co.jp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <signal.h>
#include <sys/ioctl.h>

/* x11 includes */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xmd.h>
#include <X11/extensions/shape.h>

#ifdef HAVE_LIBXPM
#include <X11/xpm.h>
#include "icongrid.xpm"
#endif /* HAVE_LIBXPM */

#ifdef HAVE_LIBXFT
#include <X11/Xft/Xft.h>
#endif /* HAVE_LIBXFT */

#ifdef HAVE_LIBXINERAMA
#include <X11/extensions/Xinerama.h>
#endif /* HAVE_LIBXINERAMA */

struct osdprivate
{
    /* Private data follows */
    Display *display;
    Window win;
    GC greengc, blackgc;
    int width,height;
    int leftedge;
    int mapped;
    int currentlevel,maxlevel;
    char *name;
    Pixmap background;
    int use_icons;
    int icon;
#ifdef HAVE_LIBXPM
    Pixmap icons;
    Pixmap iconmask;
#endif
#ifdef HAVE_LIBXFT
    XftDraw *xftdraw;
    XftFont *xftfont;
    XftColor bgcolor,fgcolor;
#endif /* HAVE_LIBXFT */
#ifdef HAVE_LIBXINERAMA
    XineramaScreenInfo *xinerama;
    int xinerama_screencount;
#endif /* HAVE_LIBXINERAMA */
};

#include "osd.h"

#define MUTE_STRING "MUTE"

#define ICONS_MAPPED 25

int iconmapping[]=
{
  0, /* volume */
  1, /* Treble */
  2, /* Bass */
  7, /* Synth */
  3, /* PCM */

  0, /* Speaker */
  4, /* Line */
  5, /* Mic */
  6, /* CD */
  7, /* IMIX */

  3, /* AltPCM */
  7, /* RecLev */
  7, /* IGain */
  7, /* OGain */
  4, /* Line1 */

  4, /* Line 2 */
  4, /* Line 3 */
  4, /* Digital 1 */
  4, /* Digital 2 */
  4, /* Digital 3 */

  4, /* Phone In */
  4, /* Phone Out */
  7, /* Video */
  7, /* Radio */
  7, /* Monitor */
};

static unsigned long getcolor(struct osdcontext *osd,char *colorname)
{
    XColor color;
    XWindowAttributes winattr;
    XGetWindowAttributes(osd->priv->display,
        RootWindow(osd->priv->display,
        DefaultScreen(osd->priv->display)), &winattr);

    color.pixel = 0;
    XParseColor(osd->priv->display, winattr.colormap, colorname, &color);
    color.flags = DoRed | DoGreen | DoBlue;
    XAllocColor(osd->priv->display, winattr.colormap, &color);
    return color.pixel;
}


static void osd_setmaxlevel(struct osdcontext *osd, int maxlevel)
{
  osd->priv->maxlevel=maxlevel;
  if(osd->priv->currentlevel > osd->priv->maxlevel)
    osd->priv->currentlevel=osd->priv->maxlevel;
}


static void osd_showname(struct osdcontext *osd)
{
  if(osd->priv->name)
  {
    XDrawString(osd->priv->display, osd->priv->win,
                osd->priv->greengc, osd->priv->leftedge, 28,
                osd->priv->name, strlen(osd->priv->name));
    XFlush(osd->priv->display);
  }

#ifdef HAVE_LIBXPM
  if(osd->priv->use_icons)
  {
    int pos=(osd->priv->height-48)/2;

    XSetClipMask(osd->priv->display, osd->priv->blackgc, osd->priv->iconmask);
    XSetClipOrigin(osd->priv->display, osd->priv->blackgc,
                   pos - 48*osd->priv->icon, pos);

    XCopyArea(osd->priv->display, osd->priv->icons, osd->priv->win,
              osd->priv->blackgc, osd->priv->icon*48, 0, 48, 48, pos, pos);

    XSetClipMask(osd->priv->display, osd->priv->blackgc, None);
  }
#endif /* HAVE_LIBXPM */
}

static void osd_setname(struct osdcontext *osd,char *name,int icon)
{
  osd->priv->name=name;

  if(osd->priv->mapped)
  {
    XCopyArea(osd->priv->display, osd->priv->background, osd->priv->win,
              osd->priv->greengc, 0, 0, osd->priv->width, osd->priv->height,
              0, 0);
  }

  if(icon<ICONS_MAPPED)
    osd->priv->icon=iconmapping[icon];
  else
    osd->priv->icon=7;

  osd_showname(osd);
}


static void osd_setlevel(struct osdcontext *osd,int level,int mute)
{
  int i;
  int cells,cellf;
  XCopyArea(osd->priv->display, osd->priv->background, osd->priv->win,
	    osd->priv->blackgc, osd->priv->leftedge, 40,
	    osd->priv->width, 25, osd->priv->leftedge, 40);
  if (!mute)
  {
    XDrawString(osd->priv->display, osd->priv->win,
		osd->priv->greengc, osd->priv->leftedge, 64,
		MUTE_STRING, strlen(MUTE_STRING));
  }
  else 
  {
    osd->priv->currentlevel=level;
    if(osd->priv->currentlevel < 0)
      osd->priv->currentlevel=0;
    if(osd->priv->currentlevel > osd->priv->maxlevel)
      osd->priv->currentlevel=osd->priv->maxlevel;
    cells = (osd->priv->width-osd->priv->leftedge) / 20;
    cellf = ((osd->priv->width-osd->priv->leftedge) * osd->priv->currentlevel);
    cellf /= osd->priv->maxlevel;
    cellf /= 20;

    for (i = 0; i < cellf; i++)
      XFillRectangle(osd->priv->display, osd->priv->win,
		     osd->priv->greengc, osd->priv->leftedge + i*20, 40, 10, 25);
    for (i = cellf; i<cells ; ++i)
      XCopyArea(osd->priv->display, osd->priv->background, osd->priv->win,
		osd->priv->blackgc, osd->priv->leftedge + i*20, 40,
		10, 25, osd->priv->leftedge + i*20, 40);
  }
  XFlush(osd->priv->display);
}


static void osd_map(struct osdcontext *osd)
{
  if(osd->priv->mapped==0)
  {
    XMapRaised(osd->priv->display, osd->priv->win);
#ifdef HAVE_LIBXFT
    XftDrawRect(osd->priv->xftdraw, &osd->priv->bgcolor, 0, 0,
                osd->priv->width, osd->priv->height);
#endif /* HAVE_LIBXFT */

    XCopyArea(osd->priv->display, osd->priv->win, osd->priv->background,
              osd->priv->blackgc, 0, 0, osd->priv->width, osd->priv->height,
              0, 0);

    osd_showname(osd);

    osd->priv->mapped = 1;
  }
}


static void osd_unmap(struct osdcontext *osd)
{
  if(osd->priv->mapped)
  {
    XUnmapWindow(osd->priv->display, osd->priv->win);
    XFlush(osd->priv->display);
    osd->priv->mapped = 0;
  }
}


static void osd_dispose(struct osdcontext *osd)
{
  if(osd->priv->mapped)
    osd->hide(osd);
  free(osd);
}

#ifdef HAVE_LIBXFT
void setup_xft(struct osdcontext *osd)
{
  XftDraw *xftdraw;
  XftFont *xftfont;
  XftColor color_bg,color_fg,colortwo;
  XRenderColor colortmp;
  int screen_num=DefaultScreen(osd->priv->display);

  osd->priv->xftdraw =
    XftDrawCreate(osd->priv->display,
                 (Drawable) osd->priv->win,
                 DefaultVisual(osd->priv->display, screen_num),
                 DefaultColormap(osd->priv->display, screen_num));

  if (!osd->priv->xftdraw) return;

  colortmp.red   = 0x0;
  colortmp.green = 0x0;
  colortmp.blue  = 0x0;
  colortmp.alpha = 0x006000;
  XftColorAllocValue(osd->priv->display,
                     DefaultVisual(osd->priv->display, screen_num),
                     DefaultColormap(osd->priv->display,screen_num),
                     &colortmp,
                     &osd->priv->bgcolor);

  colortmp.red   = 0x0;
  colortmp.green = 0x0;
  colortmp.blue  = 0x0;
  colortmp.alpha = 0x00ffff;
  XftColorAllocValue(osd->priv->display,
                     DefaultVisual(osd->priv->display, screen_num),
                     DefaultColormap(osd->priv->display,screen_num),
                     &colortmp,
                     &osd->priv->fgcolor);

}
#endif /* HAVE_LIBXFT */


#ifdef HAVE_LIBXPM
int setup_xpm(struct osdcontext *osd)
{
  XpmAttributes attr;

  attr.exactColors = 0;
  attr.alloc_close_colors = 1;
  attr.closeness = 30000;
  attr.valuemask |= (XpmExactColors | XpmAllocCloseColors | XpmCloseness);
  if (XpmCreatePixmapFromData
     (osd->priv->display,DefaultRootWindow(osd->priv->display), icongrid_xpm,
     &osd->priv->icons, &osd->priv->iconmask, NULL) != XpmSuccess)
  {
     printf("Cannot allocate colors for the knob pixmap!\n");
     return(0);
  }
  return(1);
}
#endif /* HAVE_LIBXPM */


struct osdcontext *osd_create(int widthoffset, int bottomedge, int screen)
{
  struct osdcontext *osd;
  struct osdprivate *osdp;
  XGCValues gcval;
  GC gc;
  XSizeHints sizehints;
  XSetWindowAttributes xattributes;
  int width, screen_num;
  XFontStruct *fs = NULL;
#ifdef HAVE_LIBXINERAMA
  int i;
#endif /* HAVE_LIBXINERAMA */

  if(!(osd=malloc(sizeof(struct osdcontext)))) {
    fprintf(stderr, "Can't allocate osdcontext!\n");
    return(NULL);
  }

  if(!(osdp=malloc(sizeof(struct osdprivate)))) {
    fprintf(stderr, "Can't allocate osdcontext!\n");
    return(NULL);
  }

  osd->priv=osdp;

  osdp->mapped=0;
  osdp->maxlevel=100;
  osdp->currentlevel=0;
  osdp->height = 72;

  osd->dispose=osd_dispose;
  osd->setmaxlevel=osd_setmaxlevel;
  osd->setlevel=osd_setlevel;
  osd->setname=osd_setname;
  osd->show=osd_map;
  osd->hide=osd_unmap;

  if (!(osd->priv->display = XOpenDisplay(NULL))) {
    fprintf(stderr, "Unable to open display\n");
    exit(EXIT_FAILURE);
  }

#ifdef HAVE_LIBXINERAMA
  if (XineramaQueryExtension(osdp->display,&i,&i) && XineramaIsActive(osdp->display))
  {
    osdp->xinerama = XineramaQueryScreens(osdp->display,&osdp->xinerama_screencount);
    if (screen > osdp->xinerama_screencount)
    {
      fprintf(stderr, "Xinerama screen out of range\n");
      exit(EXIT_FAILURE);
    }
    for (i = 0; i < osdp->xinerama_screencount; i++)
    {
      if (osdp->xinerama[i].screen_number==screen)
      {
        width = osdp->width = osdp->xinerama[i].width - widthoffset;
        sizehints.x = (((osdp->xinerama[i].width - width) / 2) + osdp->xinerama[i].x_org);
      }
    }
  } else { /* Xinerama is not available */
    width = osdp->width = DisplayWidth(osdp->display,
                         DefaultScreen(osdp->display)) - widthoffset;
    sizehints.x = (DisplayWidth(osd->priv->display, 0) - width) / 2;
  }
#else /* HAVE_LIBXINERAMA */
  width = osdp->width = DisplayWidth(osdp->display,
                          DefaultScreen(osdp->display)) - widthoffset;

  sizehints.x = (DisplayWidth(osd->priv->display, 0) - width) / 2;
#endif /* HAVE_LIBXINERAMA */

  screen_num = DefaultScreen(osdp->display);

  sizehints.flags = USSize | USPosition;

  sizehints.y = (DisplayHeight(osd->priv->display, 0)
                - (osd->priv->height+bottomedge));

  sizehints.width = osd->priv->width;
  sizehints.height = osd->priv->height;
  xattributes.save_under = True;
  xattributes.override_redirect = True;
  xattributes.cursor = None;

  osd->priv->win = XCreateWindow(osd->priv->display,
                           DefaultRootWindow(osd->priv->display),
                           sizehints.x,sizehints.y,
                           osd->priv->width,osd->priv->height,0,
                           CopyFromParent, // depth
                           CopyFromParent, // class
                           CopyFromParent, // visual
                           0,              // valuemask
                           0);    // attributes

  XSetWMNormalHints(osd->priv->display, osd->priv->win, &sizehints);
  XChangeWindowAttributes(osd->priv->display, osd->priv->win,
                          CWSaveUnder | CWOverrideRedirect, &xattributes);
  XStoreName(osd->priv->display, osd->priv->win, "osd");

	fs = XLoadQueryFont(osd->priv->display,
         "-misc-fixed-medium-r-normal--36-*-75-75-c-*-iso8859-*");


  if (fs == NULL) {
    fprintf(stderr, "Trying alternate font\n");
    fs = XLoadQueryFont(osd->priv->display,
           "-sony-fixed-medium-r-normal--24-*-100-100-c-*-iso8859-*");
    if (fs == NULL) {
      fprintf(stderr, "Trying \"fixed\" font\n");
      fs = XLoadQueryFont(osd->priv->display,"fixed");
      if (fs == NULL) {
        fprintf(stderr, "Your X server is probably broken\n");
        return(NULL);
      }
    }
  }

  gcval.foreground = getcolor(osd,"green");
  gcval.background = getcolor(osd,"black");
  gcval.graphics_exposures = 0;

  osd->priv->greengc = XCreateGC(osd->priv->display, osd->priv->win,
      GCForeground | GCBackground | GCGraphicsExposures, &gcval);

  gcval.foreground = getcolor(osd,"black");
  osd->priv->blackgc = XCreateGC(osd->priv->display, osd->priv->win,
      GCForeground | GCBackground | GCGraphicsExposures, &gcval);

  XSetFont(osd->priv->display, osd->priv->greengc, fs->fid);

#ifdef HAVE_LIBXFT
  setup_xft(osd);
#endif

  osd->priv->use_icons=0;
#ifdef HAVE_LIBXPM
  osd->priv->leftedge=osd->priv->height;
  if(!(osd->priv->use_icons=setup_xpm(osd)))
    osd->priv->leftedge=20;
#else
  osd->priv->leftedge=20;
#endif

  osd->priv->background = XCreatePixmap(osd->priv->display,
      DefaultRootWindow(osd->priv->display),
      osd->priv->width,osd->priv->height,
      DefaultDepth(osd->priv->display,screen_num));

  return(osd);
}

